from moexalgo import Futures, Market


def test_futures_ticker_futoi():
    SiU4 = Futures('SiU4')
    it = SiU4.futoi(start="2024-04-22", end='2024-04-22', use_dataframe=False)
    assert next(it)
    assert next(it)

    cnt = 0
    for _ in it:
        cnt += 1
        assert cnt < 9000, "Infinity loop"
    assert cnt


def test_futures_market_futoi():
    fo = Market('FO')
    it = fo.futoi(date="2024-04-22", use_dataframe=False)
    assert next(it)
    assert next(it)

    cnt = 0
    for _ in it:
        cnt += 1
        assert cnt < 9000, "Infinity loop"
    assert cnt


import pytest
from datetime import date, timedelta

from numpy.distutils.lib2def import DATA_RE

from moexalgo import session, Stock, Market, Futures

today = date.today()


def test_market(datashop_apikey):
    session.TOKEN = datashop_apikey
    eq = Market('EQ')
    assert eq == Market('shares')
    assert eq == Market('shares', 'TQBR')
    tickers = eq.tickers(use_dataframe=False)
    assert len([ticker for ticker in tickers]) > 50


def test_candles(datashop_apikey):
    session.TOKEN = datashop_apikey
    MOEX = Stock('MOEX')
    it = MOEX.candles(start=today, end=today, latest=True, use_dataframe=False)
    next(it)


def test_trades(datashop_apikey):
    session.TOKEN = datashop_apikey
    MOEX = Stock('MOEX')
    it = MOEX.trades(use_dataframe=False)
    trade = next(it)
    assert 'TRADENO' in trade


def test_tradestats(datashop_apikey):
    session.TOKEN = datashop_apikey
    EQ = Market('EQ')
    it = EQ.tradestats(date=today, use_dataframe=False)
    tradestat = next(it)
    assert tradestat.ts.date() == today
    assert 'disb' in tradestat
    assert next(it)

    MOEX = Stock('MOEX')
    it = MOEX.tradestats(start=today, end=today, latest=True, use_dataframe=False)
    tradestat = next(it)
    assert tradestat.ts.date() == today
    assert 'disb' in tradestat


def test_orderstats(datashop_apikey):
    session.TOKEN = datashop_apikey
    EQ = Market('EQ')
    it = EQ.orderstats(date=today, use_dataframe=False)
    orderstat = next(it)
    assert orderstat.ts.date() == today
    assert 'put_vwap_s' in orderstat
    assert next(it)

    MOEX = Stock('MOEX')
    it = MOEX.orderstats(start=today, end=today, latest=True, use_dataframe=False)
    orderstat = next(it)
    assert orderstat.ts.date() == today
    assert 'put_vwap_b' in orderstat


def test_obstats(datashop_apikey):
    session.TOKEN = datashop_apikey
    EQ = Market('EQ')
    it = EQ.obstats(date=today, use_dataframe=False)
    obstat = next(it)
    assert obstat.ts.date() == today
    assert 'spread_bbo' in obstat
    assert next(it)

    MOEX = Stock('MOEX')
    it = MOEX.obstats(start=today, end=today, latest=True, use_dataframe=False)
    obstat = next(it)
    assert obstat.ts.date() == today
    assert 'spread_1mio' in obstat


def test_futoi(datashop_apikey):
    session.TOKEN = datashop_apikey
    _14_day = today - timedelta(days=14)
    FO = Market('FO')
    it = FO.futoi(date=_14_day, use_dataframe=False)
    futoi = next(it)
    assert futoi.ts.date() == _14_day
    assert 'pos_short' in futoi
    assert next(it)



from __future__ import annotations

from datetime import date
from typing import Union

import pandas as pd

from moexalgo.metrics import prepare_request, dataclass_it, pandas_frame
from moexalgo.session import Session
from moexalgo.tickers import _Ticker, _resolve_ticker


class Stock(_Ticker):
    """
    Класс для работы с акциями.
    Больше информации об акциях можно найти на странице: https://moexalgo.github.io/des/supercandles/#_1

    Attributes
    ----------
    _PATH : str
        Путь к данным.
    _TYPE : str
        Тип инструмента.
    _LIMIT : int
        Количество записей в результате, если не указано другое значение (по умолчанию 25 000).
    
    Methods
    ----------
    info : Callable
        Возвращает информацию об инструменте.
    tradestats : Callable
        Возвращает метрики `TradeStat`.
    obstats : Callable
        Возвращает метрики `ObStat`.
    hi2 : Callable
        Возвращает метрики `Hi2`.

    Example
    ----------

    .. code-block:: python

        # Получение информации об акции
        >>> from moexalgo import stocks
        >>> stock = stocks.get('SBER')
        >>> info = stock.info()
        >>> print(info)

        # Получение метрик `TradeStat` по заданным параметрам (по другим статистикам аналогично)
        >>> from moexalgo import stocks
        >>> stock = stocks.get('SBER')
        >>> tradestats = stock.tradestats(start='2021-01-01', end='2021-01-10')
        >>> print(tradestats)
    """

    _PATH = 'engines/stock/markets/shares'
    _TYPE = 'eq'
    _LIMIT = 25_000

    def info(self, *fields: tuple[str], use_dataframe: bool = True) -> Union[dict, pd.DataFrame]:
        """
        Возвращает информацию об инструменте.

        Parameters
        ----------
        fields : str
            Поля для отображения.
        use_dataframe : bool
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pandas.DataFrame`, иначе `dict`.

        Returns
        ----------
        return : Union[dict, pd.DataFrame]
            Информация об инструменте.
        """
        return super().info(*fields, use_dataframe=use_dataframe)
    
    def _prepare_metric(self,
                        metric_type: str,
                        metric: str,
                        start: Union[str, date],
                        end: Union[str, date],
                        latest: bool = None,
                        offset: int = None,
                        cs: Session = None,
                        use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Подготовка метрик.

        Parameters
        ----------
        metric_type : str
            Тип инструмента.
        metric : str
            Название метрики
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (start может быть равен end, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default None.
        offset : int, optional
            Начальная позиция в последовательности записей, by default None.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        ----------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик
        """
        metrics_it = prepare_request(
            f'{metric_type}/{metric}',
            cs,
            secid=self._secid,
            from_date=start,
            till_date=end,
            latest=latest,
            offset=offset
        )
        return pandas_frame(metrics_it) if use_dataframe else dataclass_it(metrics_it)

    def tradestats(self, 
                   *, 
                   start: Union[str, date], 
                   end: Union[str, date],
                   latest: bool = None, 
                   offset: int = None,
                   cs: Session = None, 
                   use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `TradeStat` (статистику по сделкам) по заданным параметрам.
        Больше информации о метрике `TradeStat` можно найти на странице: https://moexalgo.github.io/des/supercandles/#tradestats

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default None.
        offset : int, optional
            Начальная позиция в последовательности записей, by default None.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.
        
        Returns
        ----------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `TradeStat`.
        """
        return self._prepare_metric(
            self._TYPE,
            'tradestats',
            start,
            end,
            latest,
            offset,
            cs,
            use_dataframe
        )

    def orderstats(self, 
                   *, 
                   start: Union[str, date], 
                   end: Union[str, date],
                   latest: bool = None, 
                   offset: int = None,
                   cs: Session = None, 
                   use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `OrderStat` (статистику по заявкам) по заданным параметрам.
        Больше информации о метрике `OrderStat` можно найти на странице: https://moexalgo.github.io/des/supercandles/#orderstats

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default None.
        offset : int, optional
            Начальная позиция в последовательности записей, by default None.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        ----------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `OrderStat`.
        """
        return self._prepare_metric(
            self._TYPE,
            'orderstats',
            start,
            end,
            latest,
            offset,
            cs,
            use_dataframe
        )

    def obstats(self, 
                *, 
                start: Union[str, date], 
                end: Union[str, date],
                latest: bool = None, 
                offset: int = None,
                cs: Session = None, 
                use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `ObStat` (статистику по стакану) по заданным параметрам.
        Больше информации о метрике `ObStat` можно найти на странице: https://moexalgo.github.io/des/supercandles/#obstats

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default None.
        offset : int, optional
            Начальная позиция в последовательности записей, by default None.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        ----------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `ObStat`.
        """
        return self._prepare_metric(
            self._TYPE,
            'obstats',
            start,
            end,
            latest,
            offset,
            cs,
            use_dataframe
        )

    def hi2(self, 
            *, 
            start: Union[str, date], 
            end: Union[str, date],
            latest: bool = None, 
            offset: int = None,
            cs: Session = None, 
            use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `Hi2` (индекс рыночной концентрации) по заданным параметрам.
        Чтобы узнать больше о метрике `Hi2`, посетите страницу: https://moexalgo.github.io/des/hi2/

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default None.
        offset : int, optional
            Начальная позиция в последовательности записей, by default None.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        ----------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `Hi2`
        """
        return self._prepare_metric(
            self._TYPE,
            'hi2',
            start,
            end,
            latest,
            offset,
            cs,
            use_dataframe
        )


    def alerts(self,
            *,
            start: Union[str, date],
            end: Union[str, date],
            latest: bool = None,
            offset: int = None,
            cs: Session = None,
            use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает MegaAlert (оповещение об аномальной рыночной активности) по заданным параметрам.

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default None.
        offset : int, optional
            Начальная позиция в последовательности записей, by default None.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        ----------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame`
        """
        return self._prepare_metric(
            self._TYPE,
            'alerts',
            start,
            end,
            latest,
            offset,
            cs,
            use_dataframe
        )



def get(name: str) -> Stock:
    """
    Возвращает объект для работы с акцией.

    Parameters
    ----------
    name : str
        Название акции.

    Returns
    ----------
    return : Stock
        Объект для работы с акцией.

    Example
    ----------
    .. code-block:: python

        # Получение информации об акции
        >>> from moexalgo import stocks
        >>> stock = stocks.get('ABIO')
        >>> info = stock.info()
        >>> print(info)
    """
    return Stock(name)



import pytest
from moexalgo import Ticker, Market, Stock, Index, Futures, Currency
from moexalgo.models.common import Candle


def test_tickers_creation():
    eq = Market('EQ')
    moex = Ticker('MOEX')
    assert isinstance(moex, Stock)
    assert moex._boardid == eq._boardid == 'TQBR'
    assert moex._market == eq

    ndx = Market('index')
    imoex = Ticker('IMOEX')
    assert isinstance(imoex, Index)
    assert imoex._boardid == ndx._boardid == 'SNDX'
    assert imoex._market == ndx

    fut = Market('futures')
    akm4 = Ticker('AKM4')
    assert isinstance(akm4, Futures)
    assert akm4._boardid == fut._boardid == 'RFUD'
    assert akm4._market == fut

    riz2 = Ticker('RIZ2')
    assert isinstance(riz2, Futures)
    assert riz2._boardid == fut._boardid == 'RFUD'
    assert riz2._market == fut

    cur = Market('currency')
    cny = Ticker('CNY000000TOD')
    assert isinstance(cny, Currency)
    assert cny._boardid == cur._boardid == 'CETS'
    assert cny._market == cur


def test_tickers_iter():
    moex = Ticker('MOEX')
    it = moex.candles(start='2024-01-10', end='2024-01-10', use_dataframe=False)
    assert isinstance(next(it), Candle)
    next(it)

    it = moex.tradestats(start='2024-01-10', end='2024-01-10', use_dataframe=False)
    assert isinstance(next(it), object)
    next(it)


if __name__ == '__main__':
    pytest.main()



from dataclasses import dataclass
from datetime import datetime, date, time
from decimal import Decimal


@dataclass
class Securities:
    """ Элемент блока данных `securities`

    Attributes
    ----------
    name: str
        Наименоваание
    decimals: int
        Точность, знаков после запятой
    shortname: str
        Наименование индекса
    annualhigh: Decimal
        Максимальное значение индекса за год
    annuallow: Decimal
        Минимальное значение индекса за год
    currencyid: str
        Валюта расчета
    calcmode: str
        Принцип расчета
    """
    name: str
    decimals: int
    shortname: str
    annualhigh: Decimal
    annuallow: Decimal
    currencyid: str
    calcmode: str


@dataclass
class MarketData:
    """ Элемент блока данных `marketdata`

    Attributes
    ----------
    lastvalue: Decimal
        Последнее значение индекса предыдущего торгового дня
    openvalue: Decimal
        Открытие текущего торгового дня, пунктов
    currentvalue: Decimal
        Текущее значение
    lastchange: Decimal
        Изменение текущего значения к значению предыдущего торгового дня, пунктов
    lastchangetoopenprc: Decimal
        Изменение к открытию текущего торгового дня, %
    lastchangetoopen: Decimal
        Изменение к открытию текущего торгового дня, пунктов
    updatetime: time
        Время последнего обновления
    lastchangeprc: Decimal
        Изменение текущего значения к значению предыдущего торгового дня, %
    valtoday: Decimal
        Объем торгов по бумагам из базы расчета за день, руб
    monthchangeprc: Decimal
        Изменение с начала календарного месяца, %
    yearchangeprc: Decimal
        Изменение с начала календарного года
    seqnum: int
        номер обновления (служебное поле
    systime: datetime
        Время загрузки данных системой
    time: time
        Время
    valtoday_usd: Decimal
        Объем торгов по бумагам из базы расчета за день, дол. США
    lastchangebp: Decimal
        Изменение текущего значения к значению предыдущего торгового дня, базисных пунктов
    monthchangebp: Decimal
        Изменение с начала календарного месяца, базисных пунктов
    yearchangebp: Decimal
        Изменение с начала календарного года, базисных пунктов
    capitalization: Decimal
        Капитализация бумаг, входящих в индекс, руб
    capitalization_usd: Decimal
        Капитализация бумаг, входящих в индекс, в дол. США
    high: Decimal
        Максимальное значение индекса за сессию
    low: Decimal
        Минимальное значение индекса за сессию
    tradedate: date
        Дата торгов
    tradingsession: str
        Торговая сессия
    voltoday: Decimal
        Объем сделок. Для товарных индексов – в тоннах
    """
    lastvalue: Decimal
    openvalue: Decimal
    currentvalue: Decimal
    lastchange: Decimal
    lastchangetoopenprc: Decimal
    lastchangetoopen: Decimal
    updatetime: time
    lastchangeprc: Decimal
    valtoday: Decimal
    monthchangeprc: Decimal
    yearchangeprc: Decimal
    seqnum: int
    systime: datetime
    time: time
    valtoday_usd: Decimal
    lastchangebp: Decimal
    monthchangebp: Decimal
    yearchangebp: Decimal
    capitalization: Decimal
    capitalization_usd: Decimal
    high: Decimal
    low: Decimal
    tradedate: date
    tradingsession: str
    voltoday: Decimal



import typing as t

from moexalgo.session import Session, AUTH_CERT
from moexalgo.utils import result_deserializer


def _get_sections(path: str, *sections: str,
                  session=None, **options: dict) -> t.Optional[t.Iterator[dict]]:
    with Session(session or Session(auth_cert=AUTH_CERT)) as client:
        items = client.get_objects(path, lambda data: result_deserializer(data, *sections), **options)
        return items


def get_secid_info_and_boards(secid: int) -> tuple[dict[str, t.Any], dict[str, t.Any]]:
    with Session(Session(auth_cert=AUTH_CERT)) as client:
        sections = client.get_objects(f'securities/{secid}',
                                      lambda data: result_deserializer(data, 'description', 'boards'))
        return (
            dict((item.pop('name'), item) for item in sections['description']),
            dict((item.get('boardid'), item) for item in sections['boards'])
        )



from moexalgo import Ticker


def test_shares_ticker():
    UWGN = Ticker('UWGN')
    it = UWGN.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



from __future__ import annotations

import re
from typing import Union

from .tickers import _Ticker, _resolve_ticker

try:
    from .__version__ import version as __version__
except ImportError:
    pass

from .requests import get_secid_info_and_boards

from .market import Market
from .tickers import _Ticker
from .stocks import Stock
from .indices import Index
from .currency import Currency
from .futures import Futures
from .utils import CandlePeriod

AnyTickers = Union['Stock', 'Index', 'Currency', 'Futures']


def Ticker(secid: str, boardid: str = None) -> AnyTickers:
    """
    Получение объекта финансового инструмента по тикеру и типу рынка.

    Этот метод позволяет получить объект `Index` или `Stock`, содержащий информацию
    о финансовом инструменте, идентифицируемом по тикеру и, опционально, типу рынка.

    Parameters
    ----------
    secid : str
        Тикер финансового инструмента, например "GAZP" для акций Газпрома.
    boardid : str, optional
        Идентификатор рынка, на котором торгуется инструмент,
        например "TQBR" для основного рынка акций на Московской бирже.

    Notes
    -----
    По умолчанию значение `None` для `boardid` означает использование стандартного рынка.

    Returns
    -------
    return : Union[Index, Stock]
        Объект класса Index или Stock, содержащий информацию о запрашиваемом финансовом инструменте.

    Raises
    ------
    LookupError
        Исключение возникает, если тикер не найден на указанном рынке.

    Example
    -------
    .. code-block:: python

        # Получение информации об акции
        >>> try:
        >>>     instrument = Ticker("GAZP", "TQBR")
        >>>     print(instrument)
        >>> except LookupError:
        >>>     print("Тикер не найден.")
    """
    if info := _resolve_ticker(secid, boardid):
        secid, boardid, market, engine, *args = info
        allowed_tickers = dict((item._PATH, item) for item in [Currency, Futures, Index, Stock])
        if allowed_ticker := allowed_tickers.get(f'engines/{engine}/markets/{market}'):
            return allowed_ticker(*info)
    raise LookupError(f"Cannot found ticker: `{secid}`")



import pytest
from moexalgo import Market


def test_markets_creation():
    eq = Market('EQ')
    assert eq == Market('EQ')
    assert eq == Market('shares')
    assert eq == Market('shares', 'TQBR')

    tickers = eq.tickers(use_dataframe=False)
    assert len([ticker for ticker in tickers]) > 50


    it = eq.tradestats(date="2024-01-10", use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


    index = Market('index')
    assert index != eq
    assert index == Market('index', 'SNDX')


def test_market_eq():
    eq = Market('shares')
    assert eq._pref == 'eq'
    it = eq.tradestats(date="2024-01-10", use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



def test_market_fx():
    fx = Market('currency')
    tickers = fx.tickers(use_dataframe=False)
    assert len([ticker for ticker in tickers]) > 10


    assert fx._pref == 'fx'
    it = fx.tradestats(date="2024-04-22", use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



def test_market_fo():
    fo = Market('futures')
    tickers = fo.tickers(use_dataframe=False)
    assert len([ticker for ticker in tickers]) > 10

    assert fo._pref == 'fo'
    it = fo.tradestats(date="2024-04-22", use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


def test_futoi():
    fo = Market('FO')
    it = fo.tradestats(date='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


    it = fo.futoi(date='2024-04-08', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


if __name__ == '__main__':
    pytest.main()



import pytest

from moexalgo import Ticker, Market


def test_market_eq():
    eq = Market('shares')
    it = eq.alerts(date="2024-09-23", use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


def test_shares_ticker():
    SBER = Ticker('SBER')
    it = SBER.alerts(start='2024-09-23', end='2024-09-23', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


# def test_currency_ticker():
#     cny000000tod = Ticker('CNY000000TOD')
#     it = cny000000tod.alerts(start='2024-04-22', end='2024-04-22', use_dataframe=False)
#     assert isinstance(next(it), object)



# DEPRECATED!
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal


@dataclass
class TradeStat:
    """ Метрика "Статистика торгов"

    Attributes
    ----------
    secid: str
        Код инструмента
    ts: datetime
        Дата/время данных
    pr_open: Decimal
        Цена открытия за период
    pr_high: Decimal
        Макс цена за период
    pr_low: Decimal
        Минимальная цена за период
    pr_close: Decimal
        Цена закрытия за период
    pr_change: Decimal
        Изменение цены за период, %
    trades: int
        Кол-во сделок
    vol: int
        Объем бумаг, лот
    val: Decimal
        Объем бумаг, руб.
    pr_std: Decimal
        Стандартное отклонение цены
    disb: Decimal
        Соотношение объема продавец/покупатель
    pr_vwap: Decimal
        Средневзвешенная цена
    trades_b: int
        Кол-во сделок на покупку
    vol_b: int
        Объем покупок, лот
    val_b: Decimal
        Объем покупок, руб.
    pr_vwap_b: Decimal
        Цена покупки средневзвешенная
    trades_s: int
        Кол-во сделок на продажу
    vol_s: int
        Объем продаж, лот
    val_s: Decimal
        Объем продаж, руб.
    pr_vwap_s: Decimal
        Цена продажи средневзвешенная
    """
    secid: str
    ts: datetime
    pr_open: Decimal
    pr_high: Decimal
    pr_low: Decimal
    pr_close: Decimal
    pr_change: Decimal
    trades: int
    vol: int
    val: Decimal
    pr_std: Decimal
    disb: Decimal
    pr_vwap: Decimal
    trades_b: int
    vol_b: int
    val_b: Decimal
    pr_vwap_b: Decimal
    trades_s: int
    vol_s: int
    val_s: Decimal
    pr_vwap_s: Decimal


@dataclass
class OrderStat:
    """ Метрика "Статистика заявок"

    Attributes
    ----------
    secid: str
        Код инструмента
    ts: datetime
        Момент времени на который рассчитаны метрики
    put_orders: int
        Количество поставленных в стакан сделок
    put_orders_b: int
        Количество поставленных в стакан сделок на покупку
    put_orders_s: int
        Количество поставленных в стакан сделок на продажу
    put_vol: int
        Объем заявок в лотах, поставленных в стакан
    put_vol_b: int
        Объем заявок в лотах, поставленных в стакан сделок на покупку
    put_vol_s: int
        Объем заявок в лотах, поставленных в стакан сделок на продажу
    put_val: Decimal
        Объем в деньгах заявок поставленных в стакан
    put_val_b: Decimal
        Объем в деньгах заявок поставленных в стакан сделок на покупку
    put_val_s: Decimal
        Объем в деньгах заявок поставленных в стакан на продажу
    cancel_orders: int
        Количество снятых заявок
    cancel_orders_b: int
        Количество снятых заявок на покупку
    cancel_orders_s: int
        Количество снятых заявок на продажу
    cancel_vol: int
        Объем снятых заявок в лотах
    cancel_vol_b: int
        Объем снятых заявок в лотах на покупку
    cancel_vol_s: int
        Объем снятых заявок в лотах на продажу
    cancel_val: Decimal
        Объем в деньгах снятых заявок
    cancel_val_b: Decimal
        Объем в деньгах снятых заявок на покупку
    cancel_val_s: Decimal
        Объем в деньгах снятых заявок на продажу
    put_vwap_b: Decimal
        Средневзвешенная цена поставленных заявок на покупку
    put_vwap_s: Decimal
        Средневзвешенная цена поставленных заявок на продажу
    cancel_vwap_b: Decimal
        Средневзвешенная цена отмененных заявок на покупку
    cancel_vwap_s: Decimal
        Средневзвешенная цена отмененных заявок на продажу
    """
    secid: str
    ts: datetime
    put_orders: int
    put_orders_b: int
    put_orders_s: int
    put_vol: int
    put_vol_b: int
    put_vol_s: int
    put_val: Decimal
    put_val_b: Decimal
    put_val_s: Decimal
    cancel_orders: int
    cancel_orders_b: int
    cancel_orders_s: int
    cancel_vol: int
    cancel_vol_b: int
    cancel_vol_s: int
    cancel_val: Decimal
    cancel_val_b: Decimal
    cancel_val_s: Decimal
    put_vwap_b: Decimal
    put_vwap_s: Decimal
    cancel_vwap_b: Decimal
    cancel_vwap_s: Decimal


@dataclass
class ObStat:
    """ Метрика "Стакан стакана заявок"

    Attributes
    ----------
    secid: str
        Код инструмента
    ts: datetime
        Момент времени на который рассчитаны метрики
    spread_bbo: Decimal
        Спред между лучшей ценой покупки и продажи
    spread_lv10: Decimal
        Спред между 10ым уровнем цен
    spread_1mio: Decimal
        Спред на 1 млн руб
    levels_b: int
        Кол-во уровней цен в стакане (покупка)
    levels_s: int
        Кол-во уровней цен в стакане (продажа)
    vol_b: int
        Совокупный объем заявок в стакане на всех уровнях (покупка)
    vol_s: int
        Совокупный объем заявок в стакане на всех уровнях (продажа)
    val_b: Decimal
        Совокупный объем заявок в стакане на всех уровнях (покупка), руб
    val_s: Decimal
        Совокупный объем заявок в стакане на всех уровнях (продажа), руб
    imbalance_vol_bbo: int
        Дисбаланс объема по лучшим ценам
    imbalance_val_bbo: Decimal
        Дисбаланс объема (руб) по лучшим ценам
    imbalance_vol: int
        Дисбаланс объема на всем стакане (все уровни)
    imbalance_val: Decimal
        Дисбаланс объема (руб) на всем стакане (все уровни)
    vwap_b: Decimal
        Средневзвешенная цена покупки в стакане
    vwap_s: Decimal
        Средневзвешенная цена продажи в стакане
    vwap_b_1mio: Decimal
        Цена покупки актива на 1 млн руб
    vwap_s_1mio: Decimal
        Цена продажи актива на 1 млн руб
    """
    secid: str
    ts: datetime
    spread_bbo: Decimal
    spread_lv10: Decimal
    spread_1mio: Decimal
    levels_b: int
    levels_s: int
    vol_b: int
    vol_s: int
    val_b: Decimal
    val_s: Decimal
    imbalance_vol_bbo: int
    imbalance_val_bbo: Decimal
    imbalance_vol: int
    imbalance_val: Decimal
    vwap_b: Decimal
    vwap_s: Decimal
    vwap_b_1mio: Decimal
    vwap_s_1mio: Decimal



from __future__ import annotations

from datetime import date
from typing import Union, Iterator

import pandas as pd

from moexalgo.metrics import calc_offset_limit, prepare_from_till_dates
from moexalgo.models import Candle
from moexalgo.session import Session, data_gen
from moexalgo.utils import CandlePeriod


def pandas_frame(candles_it: iter) -> pd.DataFrame:
    """
    Преобразование итератора свечей в `pd.DataFrame`.

    Parameters
    ----------
    candles_it : iter
        Итератор свечей.
    
    Returns
    -------
    return : pd.DataFrame
        Данные в формате `pd.DataFrame`.
    """

    def _make_lower_keys(dct_: dict) -> dict:
        """
        Преобразование ключей словаря в нижний регистр.

        Parameters
        ----------
        dct_ : dict
            Словарь с данными.
        
        Returns
        -------
        return : dict
            Словарь с ключами в нижнем регистре.
        """
        return {key.lower(): value for (key, value) in dct_.items()}

    return pd.DataFrame([_make_lower_keys(dct) for dct in candles_it])


def dataclass_it(candles_it: iter) -> iter[Candle]:
    """
    Преобразование итератора свечей в класс `Candle`.

    Parameters
    ----------
    candles_it : iter
        Итератор свечей.
    
    Returns
    -------
    return : iter[Candle]
        Итератор свечей в формате `Candle`.
    """
    for candle in candles_it:
        yield Candle(**candle)


def prepare_request(cs: Session, 
                    path: str, 
                    boardid: str, 
                    secid: str, 
                    *,
                    from_date: Union[str, date] = None, 
                    till_date: Union[str, date] = None,
                    period: Union[CandlePeriod, int, str] = None, 
                    offset: int = None, 
                    limit: int = None, 
                    latest: bool = False) -> Iterator[dict] | None:
    
    """
    Подготовка запроса.

    Parameters
    ----------
    cs : Session
        Сессия.
    path : str
        Путь
    boardid : str
        Идентификатор режима торгов.
    secid : str
        Идентификатор инструмента
    from_date : Union[str, date]
        Дата начала, by default None.
    till_date : Union[str, date]
        Дата окончания, by default None.
    period : Union[CandlePeriod, int, str]
        Период, by default None.
    offset : int
        Смещение, by default None.
    limit : int
        Лимит, by default None.
    latest : bool
        Последние данные, by default None.

    Returns
    -------
    return : Iterator[dict] | None
        Итератор с данными или None.

    Raises
    ------
    ValueError
        Ошибка о неправильном параметре `period`.
    TypeError
        Неверный тип для `period`.
    """
    
    def _raise_error() -> None:
        """
        Вызов ошибки.

        Raises
        ------
        ValueError
            Ошибка о неправильном параметре `period`.
        """
        raise ValueError("Неправильный параметр `period`")
    
    if isinstance(period, CandlePeriod):
        interval_seconds = period.value
    
    elif isinstance(period, int):
        if period == 1:
            interval_seconds = CandlePeriod.ONE_MINUTE.value
        elif period == 10:
            interval_seconds = CandlePeriod.TEN_MINUTES.value
        elif period == 60:
            interval_seconds = CandlePeriod.ONE_HOUR.value
        elif period == 24:
            interval_seconds = CandlePeriod.ONE_DAY.value
        elif period == 7:
            interval_seconds = CandlePeriod.ONE_WEEK.value
        elif period == 31:
            interval_seconds = CandlePeriod.ONE_MONTH.value
        else:
            _raise_error()
    
    elif isinstance(period, str):
        if period == '1min':
            interval_seconds = CandlePeriod.ONE_MINUTE.value
        elif period == '10min':
            interval_seconds = CandlePeriod.TEN_MINUTES.value
        elif period == '1h':
            interval_seconds = CandlePeriod.ONE_HOUR.value
        elif period == '1d':
            interval_seconds = CandlePeriod.ONE_DAY.value
        elif period == '1w':
            interval_seconds = CandlePeriod.ONE_WEEK.value
        elif period == '1m':
            interval_seconds = CandlePeriod.ONE_MONTH.value
        else:
            _raise_error()

    elif period is None:
        interval_seconds = CandlePeriod.TEN_MINUTES.value
    
    else:
        raise TypeError("Неверный тип для `period`")

    from_date, till_date = prepare_from_till_dates(from_date, till_date)
    
    options = {
            'from': from_date, 
            'till': till_date,
            'interval': interval_seconds
        }

    offset, limit = calc_offset_limit(offset, limit)

    if latest:
        options['iss.reverse'] = True
        limit = 1

    path = f'{path}/boards/{boardid}/securities/{secid}/candles'
    return data_gen(cs, path, options, offset, limit, 'candles')



from moexalgo.tickers import _Ticker


class Index(_Ticker):
    """
    Класс для работы с индексами.

    Attributes
    ----------
    _PATH : str
        Путь к данным.
    """
    _PATH = 'engines/stock/markets/index'


def get(name: str) -> Index:
    """
    Возвращает объект индекса.

    Parameters
    ----------
    name : str
        Название индекса.

    Returns
    -------
    Index
        Объект индекса.
    """
    return Index(name)



from dataclasses import dataclass
from datetime import time, datetime


@dataclass
class Candle:
    """Модель объекта `Свеча`

    Attributes
    ----------
    open : float
        Цена открытия свечи
    close - Цена закрытия свечи : float
    high : float
        Максимальная цена
    low : float
        Минимальная цена
    value : float
        Денежный оборот
    volume : int
        Количество
    begin : datetime
        Дата и время открытия свечи
    end : datetime
        Дата и время закрытия свечи
    """
    open: float
    close: float
    high: float
    low: float
    value: float
    volume: int
    begin: datetime
    end: datetime



from dataclasses import dataclass
from datetime import datetime, date, time
from decimal import Decimal


@dataclass
class Securities:
    """ Элемент блока данных `securities`

    Attributes
    ----------
    shortname: str
        Краткое наименование ценной бумаги
    prevprice: Decimal
        Цена последней сделки нормального периода предыдущего торгового дня
    lotsize: int
        Количество ценных бумаг в одном стандартном лоте
    facevalue: Decimal
        Номинальная стоимость одной ценной бумаги, в валюте инструмента
    status: str
        Индикатор "торговые операции разрешены/запрещены"
    boardname: str
        Режим торгов
    decimals: int
        Точность, знаков после запятой
    secname: Decimal
        Наименование финансового инструмента
    remarks: str
        Примечание
    marketcode: str
        Идентификатор рынка на котором торгуется финансовый инструмент
    instrid: str
        Группа инструментов
    sectorid: str
        Сектор (Устарело)
    minstep: Decimal
        Минимально возможная разница между ценами, указанными в заявках на покупку/продажу ценных бумаг
    prevwaprice: Decimal
        Значение оценки (WAPRICE) предыдущего торгового дня
    faceunit: str
        Код валюты, в которой выражен номинал ценной бумаги
    prevdate: date
        Дата предыдущего торгового дня
    issuesize: int
        Объем выпуска
    isin: str
        Международный идентификационный код ценной бумаги
    latname: str
        Наименование финансового инструмента на английском языке
    regnumber: str
        Номер государственной регистрации
    prevlegalcloseprice: Decimal
        Официальная цена закрытия предыдущего дня, рассчитываемая по методике ФСФР
    currencyid: str
        Валюта расчетов
    sectype: str
        Тип ценной бумаги
    listlevel: int
        Уровень листинга
    settledate: date
        Дата расчетов сделки
    """
    shortname: str
    prevprice: Decimal
    lotsize: int
    facevalue: Decimal
    status: str
    boardname: str
    decimals: int
    secname: Decimal
    remarks: str
    marketcode: str
    instrid: str
    sectorid: str
    minstep: Decimal
    prevwaprice: Decimal
    faceunit: str
    prevdate: date
    issuesize: int
    isin: str
    latname: str
    regnumber: str
    prevlegalcloseprice: Decimal
    currencyid: str
    sectype: str
    listlevel: int
    settledate: date


@dataclass
class MarketData:
    """ Элемент блока данных `marketdata`

    Parameters
    ----------
    bid: Decimal
        Лучшая котировка на покупку (Спрос)
    biddepth: int
        Объем заявок на покупку по лучшей котировке, выраженный в лотах
    offer: Decimal
        Лучшая котировка на продажу (Предложение)
    offerdepth: int
        Объем заявок на продажу по лучшей котировке, выраженный в лотах
    spread: Decimal
        Разница между лучшей котировкой на продажу и покупку (спред), руб.
    biddeptht: int
        Объем всех заявок на покупку в очереди Торговой Системы, выраженный в лотах
    offerdeptht: int
        Объем всех заявок на продажу в очереди Торговой Системы, выраженный в лотах
    open: Decimal
        Цена первой сделки
    low: Decimal
    	Минимальная цена сделки
    high: Decimal
        Максимальная цена сделки
    last: Decimal
        Цена последней сделки
    lastchange: Decimal
        Изменение цены последней сделки к цене предыдущей сделки, рублей
    lastchangeprcnt: Decimal
        Изменение цены последней сделки к цене предыдущей сделки, %
    qty: int
        Объем последней сделки, в лотах
    value: Decimal
        Объем последней сделки, в руб.
    value_usd: Decimal
        Объем последней сделки, дол. США
    waprice: Decimal
        	Средневзвешенная цена
    lastcngtolastwaprice: Decimal
        Изменение цены последней сделки к средневзвешенной цене, рублей
    waptoprevwapriceprcnt: Decimal
        Изменение средневзвешенной цены относительно средневзвешенной цены предыдущего торгового дня, %
    waptoprevwaprice: Decimal
        Изменение средневзвешенной цены к средневзвешенной цене предыдущего торгового дня, рублей
    closeprice: Decimal
        Цена послеторгового периода
    marketpricetoday: Decimal
        Рыночная цена по результатам торгов сегодняшнего дня, за одну ценную бумагу
    marketprice: Decimal
        Рыночная цена ценной бумаги по результатам торгов предыдущего дня, за одну ценную бумагу
    lasttoprevprice: Decimal
        Изменение цены последней сделки к последней цене предыдущего дня, %
    numtrades: int
        Количество сделок за торговый день
    voltoday: int
        Объем совершенных сделок, выраженный в единицах ценных бумаг
    valtoday: int
        Объем совершенных сделок, в валюте расчетов
    valtoday_usd: int
        Объем заключенных сделок, дол. США
    etfsettleprice: Decimal
        Расчетная стоимость акции/пая иностранного биржевого инвестиционного фонда
    tradingstatus: str
        Индикатор состояния торговой сессии по инструменту
    updatetime: time
        Время последнего обновления
    lastbid: Decimal
        Лучшая котировка на покупку на момент завершения нормального периода торгов
    lastoffer: Decimal
        Лучшая котировка на продажу на момент завершения нормального периода торгов
    lcloseprice: Decimal
        Официальная цена закрытия, рассчитываемая по методике ФСФР как средневзвешенная цена сделок за последние 10 минут торговой сессии, включая сделки послеторгового периода
    lcurrentprice: Decimal
        Официальная текущая цена, рассчитываемая как средневзвешенная цена сделок заключенных за последние 10 минут
    marketprice2: Decimal
        Рыночная цена 2, рассчитываемая в соответствии с методикой ФСФР
    numbids: int
        Количество заявок на покупку в очереди Торговой системы
    numoffers: int
        Количество заявок на продажу в очереди Торговой системы
    change: Decimal
        Изменение цены последней сделки по отношению к цене последней сделки предыдущего торгового дня
    time: time
        Время заключения последней сделки
    highbid: Decimal
        Наибольшая цена спроса в течение торговой сессии
    lowoffer: Decimal
        Наименьшая цена предложения в течение торговой сессии
    priceminusprevwaprice: Decimal
        Цена последней сделки к оценке предыдущего дня
    openperiodprice: Decimal
        Цена предторгового периода
    seqnum: int
        номер обновления (служебное поле)
    systime: datetime
        Время загрузки данных системой
    closingauctionprice: Decimal
        Цена послеторгового аукциона.
        В течение аукциона отображает ожидаемую цену аукциона с учетом всех зарегистрированных на текущий момент заявок.
        После завершения аукциона отображает цену состоявшегося аукциона.
    closingauctionvolume: Decimal
        Количество в сделках послеторгового аукциона.
        Ожидаемое (состоявшееся) количество лотов в сделках по указанной цене послеторгового аукциона.
    issuecapitalization: Decimal
        Текущая капитализация акции
    issuecapitalization_updatetime: time
        Время обновления капитализации
    etfsettlecurrency: str
        Валюта расчетной стоимости акции/пая иностранного биржевого инвестиционного фонда
    valtoday_rur:
        Объем совершенных сделок, рублей
    tradingsession: str
        Торговая сессия
    """
    bid: Decimal
    biddepth: int
    offer: Decimal
    offerdepth: int
    spread: Decimal
    biddeptht: int
    offerdeptht: int
    open: Decimal
    low: Decimal
    high: Decimal
    last: Decimal
    lastchange: Decimal
    lastchangeprcnt: Decimal
    qty: int
    value: Decimal
    value_usd: Decimal
    waprice: Decimal
    lastcngtolastwaprice: Decimal
    waptoprevwapriceprcnt: Decimal
    waptoprevwaprice: Decimal
    closeprice: Decimal
    marketpricetoday: Decimal
    marketprice: Decimal
    lasttoprevprice: Decimal
    numtrades: int
    voltoday: int
    valtoday: int
    valtoday_usd: int
    etfsettleprice: Decimal
    tradingstatus: str
    updatetime: time
    lastbid: Decimal
    lastoffer: Decimal
    lcloseprice: Decimal
    lcurrentprice: Decimal
    marketprice2: Decimal
    numbids: int
    numoffers: int
    change: Decimal
    time: time
    highbid: Decimal
    lowoffer: Decimal
    priceminusprevwaprice: Decimal
    openperiodprice: Decimal
    seqnum: int
    systime: datetime
    closingauctionprice: Decimal
    closingauctionvolume: Decimal
    issuecapitalization: Decimal
    issuecapitalization_updatetime: time
    etfsettlecurrency: str
    valtoday_rur: int
    tradingsession: str


@dataclass
class OrderBookItem:
    """Активные заявки

    Attributes
    ----------
    secid : str
        Код инструмента
    boardid : str
        Код режима торгов
    buysell : str
        Направленность заявки, приведшая к заключению сделки (покупка\продажа)
    price : float
        Цена котировки
    quantity : int
        Объем котировки, лотов
    seqnum : int
        номер обновления (служебное поле
    updatetime : time
        Время последнего обновления
    decimals : int
        Точность, знаков после запятой
    """
    secid: str
    boardid: str
    buysell: str
    price: float
    quantity: int
    seqnum: int
    updatetime: time
    decimals: int



from datetime import date

import pandas as pd
import pytest

from moexalgo import Ticker
from moexalgo.futures import Futures


def test_candles_futures():
    AKM4 = Futures('AKM4', 'RFUD')
    it = AKM4.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)

    it = AKM4.candles(start='2024-04-22', end='2024-04-22', use_dataframe=True)
    assert isinstance(it, pd.DataFrame)

    it = AKM4.tradestats(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


def test_futures_ticker():
    AKM4 = Ticker('AKM4')
    it = AKM4.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


def test_delisted_shares():
    RIZ2 = Futures('RIZ2', 'RFUD')
    assert RIZ2.delisted
    it = RIZ2.tradestats(start='2022-08-02', end='2022-08-02', use_dataframe=False)
    assert next(it)
    assert next(it)

    it = RIZ2.futoi(start='2022-08-02', end='2022-08-02', use_dataframe=False)
    assert next(it)
    assert next(it)


def test_futures_tradestats():
    AKM4 = Futures('AKM4', 'RFUD')
    AKM4.tradestats(start=date(2024, 4, 22), end='2024-04-22')
    AKM4.tradestats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    AKM4.tradestats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.tradestats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.tradestats(start=None, end=None)


def test_futures_obstats():
    AKM4 = Futures('AKM4', 'RFUD')
    AKM4.obstats(start=date(2024, 4, 22), end='2024-04-22')
    AKM4.obstats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    AKM4.obstats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.obstats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.obstats(start=None, end=None)


def test_futures_hi2():
    AKM4 = Futures('AKM4', 'RFUD')
    AKM4.hi2(start=date(2024, 4, 22), end='2024-04-22')
    AKM4.hi2(start=date(2024, 4, 22), end=date(2024, 4, 22))
    AKM4.hi2(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.hi2(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.hi2(start=None, end=None)


def test_futures_futoi():
    AKM4 = Futures('AKM4', 'RFUD')
    AKM4.futoi(start=date(2024, 4, 22), end='2024-04-22')
    AKM4.futoi(start=date(2024, 4, 22), end=date(2024, 4, 22))
    AKM4.futoi(start='2024-04-22', end='2024-04-22')
    # with pytest.raises(Exception):
    #     AKM4.futoi(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        AKM4.futoi(start=None, end=None)



from __future__ import annotations

from datetime import date, datetime
from typing import Union

import pandas as pd

from moexalgo.session import Session, data_gen
from moexalgo.utils import ISSTickerParamException, ISSDateParamException


def pandas_frame(metrics_it: iter[dict]) -> pd.DataFrame:
    """
    Трансформация данных из итератора в `pd.DataFrame`.

    Parameters
    ----------
    metrics_it : iter[dict]
        Итератор с данными.

    Returns
    -------
    return : pd.DataFrame
        Таблица с данными.
    """

    def normalize_row(row: dict) -> dict:
        """
        Нормализация строки.

        Parameters
        ----------
        row : dict
            Строка данных в формате `dict`.

        Returns
        -------
        return : dict
            Нормализованная строка.
        """
        return dict(
            ticker=row.pop('secid', row.pop('ticker', None)), 
            **{key.lower(): value for key, value in row.items()}
        )

    return pd.DataFrame([normalize_row(row) for row in metrics_it])


class DCls(dict):
    """
    Класс для работы с данными как с атрибутами.

    Parameters
    ----------
    data : dict
        Словарь с данными.
    """
    def __init__(self, **data) -> None:
        super().__init__(**data)

    def __getattr__(self, name):
        if name.startswith('_'):
            return self.__getattribute__(name)
        
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"object has no attribute '{name}'")


def dataclass_it(metrics_it: iter[dict]) -> iter[DCls]:
    """
    Трансформация данных из итератора в класс `DCls`.

    Parameters
    ----------
    metrics_it : iter[dict]
        Итератор с данными.

    Returns
    -------
    return : iter[DCls]
        Итератор с данными.
    """
    for data in metrics_it:
        data['ts'] = datetime.combine(data.pop('tradedate'), data.pop('tradetime'))
        data.pop('SYSTIME', data.pop('systime', None))
        yield DCls(**data)


def calc_offset_limit(offset: int = None,
                      limit: int = None,
                      min_limit: int = 1,
                      standart_limit: int = 10_000,
                      max_limit: int = 50_000,
                      min_offset: int = 0) -> tuple[int, int]:
    """
    Вычисление смещения и лимита.

    Parameters
    ----------
    offset : int, optional
        Смещение относительно начала, by default None.
    limit : int, optional
        Лимит данных, by default None.
    
    Returns
    -------
    return : tuple[int, int]
        Кортеж со смещением и лимитом.
    """
    offset = offset or min_offset
    if limit != -1:
        limit = limit or standart_limit

        limit = min_limit if limit < min_limit else max_limit if limit > max_limit else limit
        offset = min_offset if offset < min_offset else max_limit - 1 if offset >= max_limit else offset

    return offset, limit


def prepare_from_till_dates(from_date: Union[str, date] = None,
                            till_date: Union[str, date] = None) -> tuple[str, str]:
    """
    Подготовка дат начала и окончания.

    Parameters
    ----------
    from_date : Union[str, date], optional
        Дата начала, by default None.
    till_date : Union[str, date], optional
        Дата окончания, by default None.

    Returns
    -------
    return : tuple[str, str]
        Кортеж с датами начала и окончания.
    """
    if (from_date is None) or (till_date is None):
        raise ISSTickerParamException()

    from_date = date.fromisoformat(from_date) if isinstance(from_date, str) else (from_date or date.today())
    if not till_date:
        till_date = from_date
    elif isinstance(till_date, str):
        till_date = date.today() if till_date == 'today' else date.fromisoformat(till_date)
    
    if from_date > till_date:
        raise ISSDateParamException()
    
    return from_date.isoformat(), till_date.isoformat()


def get_metrics_path(metric: str, secid: str = None) -> str:
    """
    Получение пути к метрике.

    Parameters
    ----------
    metric : str
        Метрика.
    secid : str, optional
        Наименование инструмента, by default None.
    
    Returns
    -------
    return : str
        Путь к метрике.
    """
    path = f'datashop/algopack/{metric}'
    if secid is not None:
        path = '/'.join([path, secid.lower()])
    return path


def prepare_request(metric: str, 
                    cs: Session, 
                    *, 
                    secid: str = None, 
                    from_date: Union[str, date] = None,
                    till_date: Union[str, date] = None, 
                    latest: bool = False, 
                    offset: int = None, 
                    limit: int = None) -> iter[dict]:
    """
    Подготовка запроса к API.

    Parameters
    ----------
    metric : str
        Метрика.
    cs : Session
        Сессия.
    secid : str, optional
        Наименование инструмента, by default None.
    from_date : Union[str, date], optional
        Дата начала, by default None.
    till_date : Union[str, date], optional
        Дата окончания, by default None.
    latest : bool, optional
        Получить последние данные, by default `False`.
    offset : int, optional
        Смещение, by default None.
    limit : int, optional
        Лимит, by default None.
    
    Returns
    -------
    return : iter[dict]
        Итератор с данными.
    
    Raises
    ------
    ISSTickerParamException
        Вызывается, если secid не указан.
    ISSDateParamException
        Вызывается, если дата начала больше даты окончания.
    """
    
    from_date, till_date = prepare_from_till_dates(from_date, till_date)
    options = {'from': from_date, 'till': till_date}
    if latest:
        options['latest'] = 1

    offset, limit = calc_offset_limit(offset, limit)
    path = get_metrics_path(metric, secid)

    return data_gen(cs, path, options, offset, limit, section='data')


def prepare_market_request(metric: str,
                           cs: Session, 
                           *, 
                           secid: str = None, 
                           date_: Union[str, date] = None,
                           start: Union[str, date] = None,
                           end: Union[str, date] = None,
                           latest: bool = False, 
                           offset: int = None, 
                           limit: int = None) -> iter[dict]:
    """
    Подготовка запроса к API (для рынка).

    Parameters
    ----------
    metric : str
        Метрика.
    cs : Session
        Сессия.
    secid : str, optional
        Наименование инструмента, by default None.
    start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
    end : Union[str, date]
        Дата конца диапазона выдачи данных.
    date_ : Union[str, date], optional
        Дата, by default None.
    latest : bool, optional
        Получить последние данные, by default `False`.
    offset : int, optional
        Смещение, by default None.
    limit : int, optional
        Лимит, by default None.

    Returns
    -------
    return : iter[dict]
        Итератор с данными.
    """

    date_ = date.today() if (date_ is None) or (date_ == 'today') else date_
    if isinstance(date_, str):
        date_ = date.fromisoformat(date_)

    if isinstance(start, str):
        start = date.fromisoformat(start)

    if isinstance(end, str):
        end = date.fromisoformat(end)

    if secid is not None:
        options = {'from': start.isoformat(), 'till': end.isoformat()}
    else:
        options = {'date': date_.isoformat()}

    if latest:
        options['latest'] = 1
    
    offset, limit = calc_offset_limit(offset, limit)
    if metric.lower().startswith('fo/futoi'):
        if len(metric.lower().split("/")) == 3:
            section = 'futoi'
            path = f'analyticalproducts/futoi/securities/{metric.lower().split("/")[-1]}'
        else:
            section = 'futoi'
            path = f'analyticalproducts/futoi/securities'
    else:
        section = 'data'
        path = get_metrics_path(metric, secid)

    return data_gen(cs, path, options, offset, limit, section=section)



import decimal
import json as _json
from datetime import datetime, date, time
from enum import Enum


class ISSTickerParamException(Exception):
    """
    Исключение, возникающее при отсутствии параметра `ticker`.

    Attributes
    ----------
    message : str
        Сообщение об ошибке
    """

    def __init__(self, message: str = "The start and end parameters are required") -> None:
        self.message = message
        super().__init__(self.message)


class ISSDateParamException(Exception):
    """
    Исключение, возникающее при некорректном формате даты.

    Attributes
    ----------
    message : str
        Сообщение об ошибке
    """

    def __init__(self, message: str = "start must be less than or equal to end.") -> None:
        """
        Parameters
        ----------
        message : str
            Сообщение об ошибке
        
        Returns
        -------
        return : None
        """
        self.message = message
        super().__init__(self.message)


class CandlePeriod(Enum):
    """
    Временные интервалы для свечей.

    Attributes
    ----------
    ONE_MINUTE : int
        1 минута
    TEN_MINUTES : int
        10 минут
    ONE_HOUR : int
        1 час = 60 минут
    ONE_DAY : int
        1 день = 24 часа
    ONE_WEEK : int
        1 неделя = 7 дней
    ONE_MONTH : int
        1 месяц = 31 день
    """
    ONE_MINUTE = 1
    TEN_MINUTES = 10
    ONE_HOUR = 60
    ONE_DAY = 24
    ONE_WEEK = 7
    ONE_MONTH = 31


class RequiredImport:
    """
    Класс для импорта библиотеки, если она не была установлена.

    Attributes
    ----------
    __name : str
        Название библиотеки (модуля).
    """

    def __init__(self, name: str) -> None:
        """
        Parameters
        ----------
        name : str
            Название библиотеки.
        
        Returns
        -------
        return : None
        """
        self.__name = name

    def __getattr__(self, item: str) -> None:
        """
        Parameters
        ----------
        item : str
            Название атрибута.
        
        Returns
        -------
        return : None
        """
        raise ImportError(f'Required `{self.__name}`')


try:
    import pandas as pd
except ImportError:
    pandas = RequiredImport('pandas')


class json:
    """
    Класс для работы с JSON.

    Attributes
    ----------
    loads : Callable
        Загрузка JSON.
    JSONDecodeError : Exception
        Исключение при ошибке декодирования JSON.
    dumps : Callable
        Сохранение JSON.
    """
    loads = _json.loads
    JSONDecodeError = _json.JSONDecodeError

    @staticmethod
    def dumps(*args, **kwargs) -> str:
        """
        Parameters
        ----------
        args : Any
            Аргументы.
        kwargs : Any
            Ключевые аргументы.
        
        Returns
        -------
        return : str
            Строка JSON.
        """

        def default(obj: object) -> object:
            """
            Параметры по умолчанию для преобразования объектов в JSON.

            Parameters
            ----------
            obj : object
                Объект данных.

            Returns
            -------
            return : object
                Объект данных.
            """
            if isinstance(obj, decimal.Decimal):
                return float(str(obj))
            elif isinstance(obj, (datetime, date, time)):
                return obj.isoformat()
            
            raise TypeError

        return _json.dumps(*args, **kwargs, default=default)


def result_deserializer(data: dict, *sections, key: callable = None) -> dict:
    """
    Parameters
    ----------
    data : dict
        Слоаврь с данными от ISS.
    sections : Tuple
        Секции данных, by default ('securities', 'marketdata').
    key : Callable
        Ключевая функция, которая принимает элемент данных и возвращает ключ, by default None.

    Returns
    -------
    return : dict
        Словарь с данными.
    """
    result = dict()

    sections = sections or ('securities', 'marketdata')
    for section in sections:
        metadata = data[section]['metadata']
        columns = data[section]['columns']

        for values in data[section]['data']:
            item = item_normalizer(metadata, dict(zip(columns, values)))

            if key:
                result.setdefault(section, dict())[key(item)] = item
            else:
                result.setdefault(section, list()).append(item)
    
    return result


def item_normalizer(metadata: dict, item: dict) -> dict:
    """
    Нормализация данных.

    Parameters
    ----------
    metadata : dict
        Метаданные.
    item : dict
        Элемент данных.
    
    Notes
    -----
    `lambda s: …` - очень дорогая операция на больших данных без numpy.

    Returns
    -------
    return : dict
        Словарь с нормализованными данными.
    """
    conv = {
        'int32': lambda s: int(s) if s is not None else None,
        'int64': lambda s: int(s) if s is not None else None,
        'double': lambda s: float(s) if s is not None else None,
        'date': lambda s: date.fromisoformat(s.strip()) if (s is not None) and (s != '0000-00-00') else None,
        'datetime': lambda s: datetime.fromisoformat(s.strip()) if s is not None else None,
        'time': lambda s: time.fromisoformat(s.strip()) if s is not None else None
    }
    return dict((key, conv.get(metadata[key]['type'], str)(value)) for key, value in item.items())



from typing import Iterator

from moexalgo.metrics import calc_offset_limit
from moexalgo.session import Session, data_gen


def dataclass_it(it: Iterator) -> Iterator[dict]:
    for items in it:
        yield dict(**items)


def prepare_request(cs: Session,
                    path: str,
                    boardid: str,
                    secid: str,
                    *,
                    tradeno: int = None,
                    offset: int = None,
                    limit: int = None,
                    ) -> Iterator[dict]:
    """
    Подготовка запроса.

    Parameters
    ----------
    cs : Session
        Сессия.
    path : str
        Путь
    boardid : str
        Идентификатор режима торгов.
    secid : str
        Идентификатор инструмента
    tradeno : int, optional
            Номер сделки с которого выдаются данные, если не задано, то с начала дня.
    offset : int
        Смещение, by default None.
    limit : int
        Лимит, by default None.

    Returns
    -------
    return : Iterator[dict]
        Итератор с данными.
    """

    options = {
        'tradeno': tradeno,
    }
    offset, limit = calc_offset_limit(offset, limit)
    path = f'{path}/boards/{boardid}/securities/{secid}/trades'
    return data_gen(cs, path, options, offset, limit, 'trades')



from moexalgo import Ticker


def test_trades_shares():
    moex = Ticker('MOEX')
    it = moex.trades(use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)

    SBER = Ticker('SBER')
    it = SBER.tradestats(start='2024-05-06', end='2024-05-06', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


def test_trades_currency():
    CNY000000TOD = Ticker('CNY000000TOD')
    it = CNY000000TOD.trades(use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



from moexalgo import Ticker
from moexalgo.currency import Currency
from datetime import date
import pytest
import pandas as pd


def test_candles_currency():
    cny000000tod = Currency('CNY000000TOD', 'CETS')
    it = cny000000tod.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


    it = cny000000tod.candles(start='2024-04-22', end='2024-04-22', use_dataframe=True)
    assert isinstance(it, pd.DataFrame)

    it = cny000000tod.tradestats(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



def test_currency_ticker():
    cny000000tod = Ticker('CNY000000TOD')
    it = cny000000tod.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    

    it = cny000000tod.tradestats(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



def test_currency_tradestats():
    cny000000tod = Currency('CNY000000TOD', 'CETS')
    cny000000tod.tradestats(start=date(2024, 4, 22),  end='2024-04-22')
    cny000000tod.tradestats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    cny000000tod.tradestats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.tradestats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.tradestats(start=None, end=None)


def test_currency_orderstats():
    cny000000tod = Currency('CNY000000TOD', 'CETS')
    cny000000tod.orderstats(start=date(2024, 4, 22),  end='2024-04-22')
    cny000000tod.orderstats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    cny000000tod.orderstats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.orderstats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.orderstats(start=None, end=None)


def test_currency_obstats():
    cny000000tod = Currency('CNY000000TOD', 'CETS')
    cny000000tod.obstats(start=date(2024, 4, 22),  end='2024-04-22')
    cny000000tod.obstats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    cny000000tod.obstats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.obstats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.obstats(start=None, end=None)


def test_currency_hi2():
    cny000000tod = Currency('CNY000000TOD', 'CETS')
    cny000000tod.hi2(start=date(2024, 4, 22),  end='2024-04-22')
    cny000000tod.hi2(start=date(2024, 4, 22), end=date(2024, 4, 22))
    cny000000tod.hi2(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.hi2(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        cny000000tod.hi2(start=None, end=None)




from moexalgo import Ticker
from moexalgo.stocks import Stock
from datetime import date
import pytest
import pandas as pd


def test_candles_shares():
    ABIO = Stock('ABIO', 'TQBR')
    it = ABIO.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)
    

    it = ABIO.candles(start='2024-04-22', end='2024-04-22', use_dataframe=True)
    assert isinstance(it, pd.DataFrame)

    it = ABIO.tradestats(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


    VEON_RX = Stock('VEON-RX')
    it = VEON_RX.tradestats(start='2024-05-06', end='2024-05-06', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)


def test_shares_ticker():
    ABIO = Ticker('ABIO')
    it = ABIO.candles(start='2024-04-22', end='2024-04-22', use_dataframe=False)
    assert isinstance(next(it), object)
    assert next(it)



def test_shares_tradestats():
    ABIO = Stock('ABIO', 'TQBR')
    ABIO.tradestats(start=date(2024, 4, 22),  end='2024-04-22')
    ABIO.tradestats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    ABIO.tradestats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.tradestats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.tradestats(start=None, end=None)


def test_delisted_shares():
    MOEX = Stock('MOEX')
    assert not MOEX.delisted

    MFON = Stock('MORI', 'TQBR')
    assert MFON.delisted
    it = MFON.obstats(start='2022-04-07', end='2022-04-07', use_dataframe=False)
    assert next(it)
    assert next(it)


def test_shares_orderstats():
    ABIO = Stock('ABIO', 'TQBR')
    ABIO.orderstats(start=date(2024, 4, 22),  end='2024-04-22')
    ABIO.orderstats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    ABIO.orderstats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.orderstats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.orderstats(start=None, end=None)


def test_shares_obstats():
    ABIO = Stock('ABIO', 'TQBR')
    ABIO.obstats(start=date(2024, 4, 22),  end='2024-04-22')
    ABIO.obstats(start=date(2024, 4, 22), end=date(2024, 4, 22))
    ABIO.obstats(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.obstats(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.obstats(start=None, end=None)


def test_shares_hi2():
    ABIO = Stock('ABIO', 'TQBR')
    ABIO.hi2(start=date(2024, 4, 22),  end='2024-04-22')
    ABIO.hi2(start=date(2024, 4, 22), end=date(2024, 4, 22))
    ABIO.hi2(start='2024-04-22', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.hi2(start='2024-04-23', end='2024-04-22')
    with pytest.raises(Exception):
        ABIO.hi2(start=None, end=None)



from __future__ import annotations

import re
from datetime import date
from typing import Optional, Union

from moexalgo import session
from moexalgo.metrics import prepare_market_request, dataclass_it, pandas_frame
from moexalgo.session import Session, data_gen
from moexalgo.utils import result_deserializer, pd

_AVAILABLE = {
    'index': dict(),
    'shares': dict(),
    'selt': dict(),

    'forts': dict()
}
_ALIASES = {
    'index': ('engines/stock/markets/index', 'SNDX'),
    'shares': ('engines/stock/markets/shares', 'TQBR'),
    'stock': ('engines/stock/markets/shares', 'TQBR'),
    'stocks': ('engines/stock/markets/shares', 'TQBR'),
    'EQ': ('engines/stock/markets/shares', 'TQBR'),

    'selt': ('engines/currency/markets/selt', 'CETS'),
    'FX': ('engines/currency/markets/selt', 'CETS'),
    'currency': ('engines/currency/markets/selt', 'CETS'),

    'forts': ('engines/futures/markets/forts', 'RFUD'),
    'FO': ('engines/futures/markets/forts', 'RFUD'),
    'futures': ('engines/futures/markets/forts', 'RFUD'),
}


def market_for(secid: str, boardid: str, cs: Session = None) -> Optional[Market]:
    """
    Возвращает рынок, на котором торгуется заданный инструмент.

    Parameters
    ----------
    secid : str
        Уникальный идентификатор инструмента.
    boardid : str
        Идентификатор рынка, указывающий на специфическую торговую площадку или сегмент рынка.
    cs : Session, optional
        Клиентская сессия, если используется.

    Returns
    -------
    return : Optional[Market]
        Экземпляр класса Market, представляющий рынок, на котором торгуется заданный инструмент.
    """
    for _, boards in _AVAILABLE.items():
        if market := boards.get(boardid):
            if market.securities_for(secid, cs):
                return market


class Market:
    """
    Представление конкретного раздела биржевого рынка.

    Больше информации о FUTOI можно найти на странице: https://moexalgo.github.io/des/futoi/

    Этот класс предоставляет доступ к информации о биржевом рынке, а также позволяет
    оперировать данными конкретного рынка по его уникальному идентификатору.

    Parameters
    ----------
    name : str
        Название рынка или его символическое имя.
    boardid : str, optional
        Идентификатор рынка, указывающий на специфическую торговую площадку или сегмент рынка.
        Если не указан, класс попытается автоматически определить идентификатор на основе общих правил.

    Methods
    -------
    tickers : Callable[[Session, bool], Union[List[Dict[str, Any]], pandas.DataFrame]]
        Возвращает список всех инструментов рынка.
    
    marketdata : Callable[[Session, bool], Union[List[Dict[str, Any]], pandas.DataFrame]]
        Возвращает статистическую информацию о всех инструментах рынка.
    
    tradestats : Callable[[Session, bool], Union[Iterator[TradeStat], pandas.DataFrame]]
        Возвращает метрики `TradeStat` по заданным параметрам.
    
    orderstats : Callable[[Session, bool], Union[Iterator[OrderStat], pandas.DataFrame]]
        Возвращает метрики `OrderStat` по заданным параметрам.
    
    obstats : Callable[[Session, bool], Union[Iterator[ObStat], pandas.DataFrame]]
        Возвращает метрики `ObStat` по заданным параметрам.
    
    futoi : Callable[[Session, bool], Union[Iterator[FUTOI], pandas.DataFrame]]
        Возвращает метрики `FUTOI` по заданным параметрам.
    
    Returns
    -------
    return : Market
        Экземпляр класса Market, представляющий указанный рынок.

    Raises
    ------
    NotImplementedError
        Вызывается, если рынок с указанным названием не поддерживается.

    Example
    -------
    .. code-block:: python

        # Получение информации об акции
        >>> try:
        >>>     market_instance = Market("index", "MOEX")
        >>>     print(market_instance)
        >>>     market_instance._ensure_loaded()
        >>> except NotImplementedError:
        >>>     print("Рынок не поддерживается.")
    """
    _name: str
    _path: str
    _pref: str
    _boardid: str
    _fields: dict[str, dict[str, dict]] = None
    _values: dict[str, dict[str, dict]] = None
    _delisted: list[str] = None
    _LIMIT = 25_000

    def __new__(cls, name: str, boardid: str = None) -> Market:
        """
        Создает новый экземпляр класса Market.

        Parameters
        ----------
        name : str
            Название рынка или его символическое имя.
        boardid : str, optional
            Идентификатор рынка, указывающий на специфическую торговую площадку или сегмент рынка.
            Если не указан, класс попытается автоматически определить идентификатор на основе общих правил.
        
        Returns
        -------
        return : Market
            Экземпляр класса Market, представляющий указанный рынок.
        
        Raises
        ------
        NotImplementedError
            Вызывается, если рынок с указанным названием не поддерживается.

        Example
        -------
        .. code-block:: python
        
            # Получение информации об акции
            >>> market_instance = Market("index", "MOEX")
            >>> print(market_instance)
        """
        if '/' not in name:
            path, default_boardid = _ALIASES.get(name, (None, None))
            boardid = boardid or default_boardid
            name = path.split('/')[-1]
        else:
            path = name
            name = name.split('/')[-1]

        if name not in _AVAILABLE:
            raise NotImplementedError(f"Market {name} is not supported")

        market = _AVAILABLE.setdefault(name, dict())
        prefs = [k for k, v in _ALIASES.items() if v[0].startswith(path) and len(k) == 2]
        pref = prefs[0].lower() if prefs else None
        if boardid not in market:
            market[boardid] = super().__new__(cls)
            market[boardid]._name = name
            market[boardid]._path = path
            market[boardid]._pref = pref
            market[boardid]._boardid = boardid

        return market[boardid]

    def __repr__(self):
        return f"{self.__class__.__name__}('{self._name}/{self._boardid}')"

    def _ensure_loaded(self, cs: Session = None) -> None:
        """
        Проверяет, загружены ли данные о рынке и его инструментах.
        Если данные не загружены, загружает их.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется.

        Returns
        -------
        return : None
        """
        if self._fields is None or self._values is None:
            with Session(cs or session.default) as client:
                self._fields = client.get_objects(
                    f'{self._path}/boards/{self._boardid}/securities/columns',
                    lambda data: result_deserializer(data, key=lambda item: item['name']))

                self._values = client.get_objects(
                    f'{self._path}/boards/{self._boardid}/securities/',
                    lambda data: result_deserializer(data, key=lambda item: item['SECID']))

    def _is_delisted(self, secid: str, cs: Session = None):
        raise NotImplementedError("Deprecated")
        # self._ensure_loaded(cs)
        # if self._delisted is None:
        #     market = self._name
        #     engine = self._path.split('/')[1]
        #     "&group_by=type&group_by_filter=preferred_share&iss.only=securities&securities.columns=secid"
        #     with Session(cs or session.default) as client:
        #         data = data_gen(cs, f'securities/',
        #                         {'engine': engine, 'market': market, 'is_trading': 0,
        #                          'iss.only': 'securities', 'securities.columns': 'secid'},
        #                         0, 10000, 'securities')
        #         self._delisted = [row['secid'] for row in data]
        # return secid in self._delisted

    def _ticker_info(self, secid: str, cs: Session = None) -> dict[str, dict[str, dict]]:
        """
        Возвращает информацию о заданном инструменте.

        Parameters
        ----------
        secid : str
            Уникальный идентификатор инструмента.
        cs : Session, optional
            Клиентская сессия, если используется.

        Returns
        -------
        return : Dict[str, Any]
            Информация о заданном инструменте.
        """
        raise NotImplementedError("Deprecated")
        # self._ensure_loaded(cs)
        # marketdata = self._values['marketdata'].get(secid)
        # securities = self._values['securities'].get(secid)
        # if securities or marketdata:
        #     return dict(securities=securities, marketdata=marketdata)
        # if self._is_delisted(secid):
        #     return dict(securities=None, marketdata=None)

    def _normalize_row(self, row: dict[str, dict], fields: tuple[str]) -> dict[str, dict]:
        """
        Нормализует строку данных о статистике инструмента.

        Parameters
        ----------
        row : Dict[str, Any]
            Строка данных о статистике инструмента.
        fields : Tuple[str]
            Поля, которые необходимо оставить в строке данных.

        Returns
        -------
        return : Dict[str, Any]
            Нормализованная строка данных о статистике инструмента.
        """
        return dict(
            ticker=row['SECID'],
            **{key.lower(): value for key, value in row.items() if key in fields}
        )

    def _get_data(self,
                  option: str,
                  cs: Session = None,
                  use_dataframe: bool = True,
                  fields: tuple[str] = None) -> Union[list[dict], pd.DataFrame]:
        """
        Возвращает данные о рынке по заданному параметру.

        Parameters
        ----------
        option : str
            Параметр, по которому необходимо получить данные.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе список.
        fields : Tuple[str], optional
            Поля, которые необходимо оставить в строке данных.
        
        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            Объекты типа List или `pd.DataFrame`.
        """
        self._ensure_loaded(cs)
        if use_dataframe:
            return pd.DataFrame(
                [self._normalize_row(row, fields) for row in self._values[option].values()]
            )
        else:
            return list(self._values[option].values())

    def tickers(self, cs: Session = None, use_dataframe: bool = True) -> Union[list[dict], pd.DataFrame]:
        """
        Возвращает список всех инструментов рынка.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе список.
        
        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            Объекты типа List или `pd.DataFrame`.
        """
        fields = (
            'SHORTNAME',  # Краткое наименование
            'LOTSIZE',  # Размер лота
            'DECIMALS',  # Количество знаков после запятой
            'MINSTEP',  # Минимальный шаг цены
            'ISSUESIZE',  # Объем выпуска
            'ISIN',  # Стандартное наименование
            'REGNUMBER',  # Регистрационный номер
            'LISTLEVEL'  # Уровень листинга
        )

        return self._get_data('securities', cs, use_dataframe, fields)

    def marketdata(self, cs: Session = None, use_dataframe: bool = True) -> Union[list[dict], pd.DataFrame]:
        """ 
        Возвращает статистическую информацию о всех инструментах рынка.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе список.

        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            Объекты типа List или `pd.DataFrame`.
        """
        fields = (
            'BID',  # Лучшая цена покупки
            'OFFER',  # Лучшая цена продажи
            'BIDDEPTHT',  # Глубина стакана покупки
            'OFFERDEPTHT',  # Глубина стакана продажи
            'OPEN',  # Цена открытия
            'HIGH',  # Максимальная цена
            'LOW',  # Минимальная цена
            'LAST',  # Цена последней сделки
            'WAPRICE',  # Средневзвешенная цена
            'LASTTOPREVPRICE',  # Изменение цены последней сделки к предыдущей
            'NUMTRADES',  # Количество сделок
            'VOLTODAY',  # Объем сделок за день
            'VALTODAY',  # Объем сделок за день в валюте
            'VALTODAY_USD',  # Объем сделок за день в долларах
            'OPENPERIODPRICE',  # Цена открытия периода
            'CLOSINGAUCTIONPRICE',  # Цена закрытия аукциона
            'CLOSINGAUCTIONVOLUME',  # Объем закрытия аукциона
            'ISSUECAPITALIZATION',  # Капитализация
            'UPDATETIME',  # Время обновления
            'SYSTIME'  # Время системы
        )

        return self._get_data('marketdata', cs, use_dataframe, fields)

    def _prepare_metric(self,
                        metric: str,
                        metric_type: str = None,
                        date: Union[str, date] = None,
                        latest: bool = None,
                        offset: int = None,
                        cs: Session = None,
                        use_dataframe: bool = True,
                        limit: int = None) -> Union[iter, pd.DataFrame]:
        """
        Подготавливает запрос к рынку по заданным параметрам.

        Parameters
        ----------
        metric : str
            Метрика, по которой необходимо получить данные.
        metric_type : str, optional
            Тип метрики, by default None
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Iterator[TradeStat] | pd.DataFrame
            Итератор или `pd.DataFrame` метрик.
        """
        metrics_it = prepare_market_request(
            f'{metric_type}/{metric}',
            cs,
            date_=date,
            latest=latest,
            offset=offset,
            limit=limit or 50_000
        )
        return pandas_frame(metrics_it) if use_dataframe else dataclass_it(metrics_it)

    def tradestats(self,
                   *,
                   date: Union[str, date] = None,
                   latest: bool = None,
                   offset: int = None,
                   cs: Session = None,
                   use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """ 
        Возвращает метрики `TradeStat` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если True, то возвращает pd.DataFrame, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или pd.DataFrame метрик `TradeStat`.
        """

        return self._prepare_metric(
            'tradestats',
            self._pref,
            date,
            latest,
            offset,
            cs,
            use_dataframe
        )

    def orderstats(self,
                   *,
                   date: Union[str, date] = None,
                   latest: bool = None,
                   offset: int = None,
                   cs: Session = None,
                   use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `OrderStat` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или pd.DataFrame метрик `OrderStat`.
        """

        return self._prepare_metric(
            'orderstats',
            self._pref,
            date,
            latest,
            offset,
            cs,
            use_dataframe
        )

    def obstats(self,
                *,
                date: Union[str, date] = None,
                latest: bool = None,
                offset: int = None,
                cs: Session = None,
                use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `ObStat` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `ObStat`.
        """
        return self._prepare_metric(
            'obstats',
            self._pref,
            date,
            latest,
            offset,
            cs,
            use_dataframe
        )

    def futoi(self,
              *,
              date: Union[str, date] = None,
              # latest: bool = None,
              # offset: int = None,
              cs: Session = None,
              use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `FUTOI` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `FUTOI`.

        Raises
        ------
        NotImplementedError
            Вызывается, если `FUTOI` не поддерживается для данного рынка.
        """
        if not self._path.startswith('engines/futures/markets'):
            raise NotImplementedError("FUTOI is not implemented for this market")

        return self._prepare_metric(
            'FUTOI',
            self._pref,
            date,
            latest=None,
            offset=None,
            cs=cs,
            use_dataframe=use_dataframe,
            limit=-1
        )

    def alerts(self,
                *,
                date: Union[str, date] = None,
                latest: bool = None,
                offset: int = None,
                cs: Session = None,
                use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает MegaAlert (оповещение об аномальной рыночной активности) по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame`.
        """
        if not self._path.startswith('engines/stock/markets'):
            raise NotImplementedError
        return self._prepare_metric(
            'alerts',
            self._pref,
            date,
            latest,
            offset,
            cs,
            use_dataframe
        )


from __future__ import annotations
import typing as t

from datetime import date, datetime
import re
from typing import Union
import weakref

from moexalgo import trades
from moexalgo.requests import get_secid_info_and_boards
from moexalgo.candles import Candle, prepare_request, pandas_frame, dataclass_it
from moexalgo.market import Market
from moexalgo.metrics import prepare_market_request, dataclass_it as dict_it
from moexalgo.session import Session, data_gen
from moexalgo.utils import pd, CandlePeriod


class _Ticker:
    """
    Базовый класс для работы с инструментами.

    Attributes
    ----------
    _PATH : str
        Путь к данным.
    _secid : str
        Идентификатор инструмента.
    _boardid : str
        Идентификатор рынка.
    _r_market : weakref.ref
        Ссылка на объект рынка.

    Methods
    -------
    info : Callable
        Возвращает информацию об инструменте.
    marketdata : Callable
        Возвращает рыночную информацию и статистику об инструменте.
    candles : Callable
        Возвращает итератор свечей инструмента по заданным параметрам.
    orderbook : Callable
        Возвращает текущий стакан лучших цен.

    Example
    -------
    .. code-block:: python
    
        # Получение информации об акции
        >>> from moexalgo import tickers
        >>> ticker = tickers.get('SBER')

        >>> print(ticker)

        # Получение информации об инструменте
        >>> info = ticker.info()
        >>> print(info)

            # Получение рыночной информации и статистики об инструменте
        >>> marketdata = ticker.marketdata()
        >>> print(marketdata)

            # Получение свечей инструмента по заданным параметрам
        >>> candles = ticker.candles(start='2021-01-01', end='2021-01-10')
        >>> for candle in candles:
        ...     print(candle)

            # Получение текущего стакана лучших цен
        >>> orderbook = ticker.orderbook()
        >>> print(orderbook)
    """

    _PATH = 'Основная часть пути к API, должна быть определена в суперклассах'

    _secid: str
    _boardid: str
    _delisted: bool
    _sec_info: dict[str, t.Any] = dict()
    _board_info: dict[str, t.Any] = dict()

    def __new__(cls, secid: str, boardid: str = None, market: str = None,
                engine: str = None, description: dict = None, board_info: dict = None) -> _Ticker:
        """
        Создает новый объект инструмента.

        Parameters
        ----------
        secid : str
            Идентификатор инструмента.
        boardid : str
            Идентификатор рынка.
        market : str
            Рынок

        Returns
        -------
        _Ticker
            Объект инструмента.
        """
        if boardid is None or market is None:
            if info := _resolve_ticker(secid, boardid):
                secid, boardid, market, engine, description, board_info = info
            else:
                raise LookupError(f"Cannot found ticker: `{secid}`")

        market = Market(market, boardid)
        instance = super().__new__(cls)
        instance._secid = secid
        instance._boardid = market._boardid
        instance._r_market = weakref.ref(market)
        instance._sec_info = description
        instance._board_info = board_info
        return instance

    @property
    def delisted(self) -> bool:
        return not self._board_info['listed_from'] <= datetime.now().date() <= self._board_info['listed_till']

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self._secid}/{self._boardid}')"

    @property
    def _market(self) -> Market:
        """
        Раздел рынка в который входит инструмент.

        Returns
        -------
        Market
            Раздел рынка.
        """
        return self._r_market()

    def info(self, *fields: tuple[str], use_dataframe: bool = True) -> Union[dict, pd.DataFrame]:
        """
        Возвращает информацию об инструменте, словарь или `pd.DataFrame`

        Parameters
        ----------
        fields : tuple[str]
            Поля для отображения
        use_dataframe : bool
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе словарь.

        Returns
        -------
        return : Union[dict, pd.DataFrame]
            Информация об инструменте
        """

        if self._market:

            if info := self._market._ticker_info(self._secid):

                if securities := info.get('securities'):

                    if not fields:
                        fields = tuple(self._market._fields['securities'].keys())
                    
                    exclude_fields = ('STATUS', 'LATNAME', 'CURRENCYID', 'SECTYPE')
                    fields = tuple(filter(lambda f: f not in exclude_fields, fields))
                    securities = list(filter(lambda x: x[0] in fields, securities.items()))

                    if use_dataframe:
                        index, value = zip(*securities)
                        return pd.DataFrame(dict(value=value), index=index)
                    
                    else:
                        return dict(securities)

    def marketdata(self, *fields, use_dataframe: bool = True) -> Union[dict, pd.DataFrame]:
        """
        Возвращает рыночную информацию и статистику об инструменте.

        Parameters
        ----------
        fields : tuple[str]
            Поля для отображения
        use_dataframe : bool
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе словарь.

        Returns
        -------
        return : Union[dict, pd.DataFrame]
            Рыночная информация и статистика об инструменте.
        """

        if self._market:

            if info := self._market._ticker_info(self._secid):

                if securities := info.get('marketdata'):

                    if not fields:
                        fields = tuple(self._market._fields['marketdata'].keys())
                    
                    titles = self._market._fields['marketdata']
                    securities = [(name, titles[name]['title'], value) for (name, value) in securities.items()]
                    securities = list(filter(lambda x: x[0] in fields, securities))

                    index, title, value = zip(*securities)
                    if use_dataframe:
                        return pd.DataFrame(dict(title=title, value=value), index=index)
                    else:
                        return dict(zip(index, value))

    def candles(self, 
                *, 
                start: Union[str, date], 
                end: Union[str, date],
                period: Union[str, int, CandlePeriod] = None, 
                offset: int = 0,
                cs: Session = None, 
                latest: bool = False, 
                use_dataframe: bool = True) -> Union[iter[Candle], pd.DataFrame]:
        """
        Возвращает итератор свечей инструмента по заданным параметрам.

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        period : Union[str, int, CandlePeriod], optional
            Период свечи, by default None.

            Если `int`, то интерпретируется в следующем формате:
            - 1 - 1 минута
            - 10 - 10 минут
            - 60 - 1 час
            - 24 - 1 день
            - 7 - 1 неделя
            - 31 - 1 месяц

            Если `str`, то интерпретируется в следующем формате:
            - '1min' - 1 минута
            - '10min' - 10 минут
            - '1h' - 1 час
            - '1d' - 1 день
            - '1w' - 1 неделя
            - '1m' - 1 месяц
        
        offset : int, optional
            Начальная позиция в последовательности записей, by default 0.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default False.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter[Candle], pd.DataFrame]
            Итератор свечей.
        """

        candles_it = prepare_request(
            cs, 
            self._PATH, 
            self._boardid, 
            self._secid, 
            period=period,
            from_date=start, 
            till_date=end, 
            offset=offset,
            latest=latest
        )
        return pandas_frame(candles_it) if use_dataframe else dataclass_it(candles_it)

    def orderbook(self, cs: Session = None, use_dataframe: bool = True) -> Union[iter, pd.DataFrame]:
        """
        Возвращает текущий стакан лучших цен.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Стакан лучших цен.
        """
        if self._PATH.startswith('engines/currency/markets'):
            raise NotImplementedError("OrderBook is not implemented for currencies")
        path = f'{self._PATH}/boards/{self._boardid}/securities/{self._secid}/orderbook'
        orderbook_it = data_gen(
            cs=cs, 
            path=path, 
            options={}, 
            offset=0, 
            limit=-1,
            section='orderbook'
        )
        return pandas_frame(orderbook_it) if use_dataframe else dataclass_it(orderbook_it)

    def futoi(
            self,
            *,
            start: Union[str, date],
            end: Union[str, date],
            # latest: bool = None,
            # offset: int = None,
            cs: Session = None,
            use_dataframe: bool = True
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `FUTOI` по заданным параметрам.

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `FUTOI`.

        Raises
        ------
        NotImplementedError
            Вызывается, если `FUTOI` не поддерживается для данного рынка.
        """
        sectype = self._secid
        if self._secid not in ('USDRUBF', 'EURRUBF', 'CNYRUBF'):
            sectype = self._secid[:2]
        metrics_it = prepare_market_request(
            f'fo/futoi/{sectype}',
            cs,
            secid=self._secid,
            start=start,
            end=end,
            # latest=latest,
            # offset=offset,
            limit=-1
        )
        return pandas_frame(metrics_it) if use_dataframe else dict_it(metrics_it)

    @classmethod
    def _get_sec_info(cls, secid: str):
        if secid not in cls._sec_info:
            rv = data_gen(None, f'securities/{secid}', {}, 0, 100, section='boards')
            if found := [info for info in rv if info['is_primary'] == 1]:
                cls._sec_info[secid] = found[0]
        try:
            return cls._sec_info[secid]
        except KeyError:
            raise LookupError(f"Cannot found ticker: `{secid}`")

    def trades(self,
                *,
                tradeno: int = None,
                cs: Session = None,
                use_dataframe: bool = True) -> Union[iter[dict], pd.DataFrame]:
        """
        Возвращает итератор сделок за последний день или начиная с заданного `tradeno`.

        Parameters
        ----------
        tradeno : int, optional
            Номер сделки с которого выдаются данные, если не задано, то с начала дня.
        cs : Session, optional
            Клиентская сессия, если используется, by default None.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter[dict], pd.DataFrame]
            Итератор сделок.
        """

        trades_it = trades.prepare_request(
            cs,
            self._PATH,
            self._boardid,
            self._secid,
            tradeno=tradeno
        )
        return pandas_frame(trades_it) if use_dataframe else trades.dataclass_it(trades_it)

def _resolve_ticker(secid: str, boardid: str = None) -> tuple[str, str, str, str, dict, dict]:
    if boardid is None:
        secid, *args = re.split('[^a-zA-Z0-9-]', secid)
        if args:
            boardid = args[0]
    description, boards = get_secid_info_and_boards(secid)
    if found := [board for name, board in boards.items() if board['is_primary']]:
        board_info = found[0]
        if boardid:
            if boardid != board_info['boardid']:
                raise ValueError("Wrong `boardid`")
        boardid = board_info['boardid']
        market = board_info['market']
        engine = board_info['engine']
        return secid, boardid, market, engine, description, board_info
